#
# Copyright 2010 Frank Adelstein.  All Rights Reserved.
#
# Permission to use, copy, modify, and distribute this
# software is hereby granted without fee, provided that
# the copyright notice and permission notice are not removed.
#

#
# DM History object, use for undo
#
# maintains an array of history entries
# each entry consists of a name, an array
# of variables that were changed, an array
# of the previous values of the variables
#


#
# main public methods: do, undo, redo
#

# queue pointer

#
# a [un]do-queue: array of entries, each entry has:
#   action name, and 3 parallel arrays: array of 
#   variable reference, array of original values, array of new values
#


use strict;
package History;

sub new {
  my $class = shift;
  my $self = {};
  bless $self, $class;
  $self->initialize();
  return $self;
}


###############################
#                             #
# record action in a do-queue #
#                             #
# input:                      #
#   action name               #
#   variable reference        #
#   old value                 #
#   new value                 #
#                             #
# Side effect: adding a new   #
# entry blows away the old    #
# redo queue                  #
#                             #
###############################
sub do {
  my $self   = shift;
  my $name   = shift;
  my $ref    = shift;
  my $oldval = shift;
  my $newval = shift;

  # create a new associative array to hold the record
  my $hashref = {name     => $name,
		 variable => $ref,
		 oldval   => $oldval,
		 newval   => $newval};

  # add it to the undo queue
  my $aref = $self->{'undo-q'};
  push @$aref, $hashref;

  return;
}

##########################
#                        #
#                        #
# reset variables to     #
# previous stateand make #
# redo queue record      #
#                        #
#                        #
##########################
sub undo {
  my $self  = shift;
  my $Whash = shift;
  my $Vars  = shift;

  $self->q_do($Whash, $Vars, 'undo-q', 'redo-q', 'oldval');
  return;

}

##########################
#                        #
#                        #
# replay action from     #
# redo queue             #
#                        #
#                        #
##########################
sub redo {
  my $self  = shift;
  my $Whash = shift;
  my $Vars  = shift;

  $self->q_do($Whash, $Vars, 'redo-q', 'undo-q', 'newval');
  return;

}

############################
#                          #
# return the undo history  #
#                          #
############################
sub show {
  my $self  = shift;

  my $history;

  my $aref = $self->{'undo-q'};

  # build up the history string
  foreach (@$aref) {
    $history .= "name: " . $_->{'name'} . ", old: " . 
      $_->{'oldval'}->[0] . ", new: " . $_->{'newval'}->[0]
	. "\n";
  }

  return $history;
}


#
# internal function to set up stuff
#
sub initialize {
  my $self = shift;

  # create the undo and redo queues
  $self->{'undo-q'} = [];
  $self->{'redo-q'} = [];
  return;
}


#
# internal function to move
# stuff from a queue to variables
# to implement undo and redo
#
sub q_do {
  my $self  = shift;
  my $Whash = shift;
  my $Vars  = shift;
  my $q1    = shift;	# source queue
  my $q2    = shift;	# dest queue
  my $vname = shift;

  my $aref = $self->{$q1};

  # will return and do nothing if nothing left on stack
  my $q = pop @$aref || return;

  my $qaref = $q->{'variable'};
  my $count = scalar @$qaref;

  for (my $i = 0; $i < $count; $i++) {
    # go through and assign the values to the variables
    my $varref = $qaref->[$i];
    my $valref = $q->{$vname};
    $$varref = $valref->[$i];
  }

  # put the entry on the other queue
  $aref = $self->{$q2};
  push @$aref, $q;


  my $old = $vname;
  my $new = ($vname eq 'oldval')? 'newval' : 'oldval';

  # add "bloodied" to condition if this damage crosses the line
  if ($q->{'variable'}[0] == \$Vars->{'curHP'} && 
      $q->{$new}[0] > $Vars->{'bloodied'} &&
      $q->{$old}[0] <= $Vars->{'bloodied'}) {
    main::add_text_to_conditions ($Whash, "bloodied");
  }


  # delete "bloodied condition if this crosses above the bloodied value
  if ($q->{'variable'}[0] == \$Vars->{'curHP'} && 
      $q->{$new}[0] <= $Vars->{'bloodied'} &&
      $q->{$old}[0] > $Vars->{'bloodied'}) {
    main::delete_condition_text ($Whash, "bloodied");
  }

  main::update_UI ($Whash, $Vars);

}

1;
